import argparse
from deprecated import deprecated
import click

from GANDLF.cli import config_generator, copyrightMessage
from GANDLF.entrypoints import append_copyright_to_help
from GANDLF.utils import logger_setup


def _generate_config(config: str, strategy: str, output: str):
    config_generator(config, strategy, output)
    print("Finished.")


@click.command()
@click.option(
    "--config",
    "-c",
    help="Path to base config.",
    required=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
)
@click.option(
    "--strategy",
    "-s",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    required=True,
    help="Config creation strategy in a yaml format.",
)
@click.option(
    "--output",
    "-o",
    required=True,
    type=click.Path(file_okay=False, dir_okay=True),
    help="Path to output directory.",
)
@click.option(
    "--log-file",
    type=click.Path(),
    default=None,
    help="Output file which will contain the logs.",
)
@append_copyright_to_help
def new_way(config, strategy, output, log_file):
    """Generate multiple GaNDLF configurations based on a single baseline GaNDLF for experimentation."""

    logger_setup(log_file)
    _generate_config(config, strategy, output)


# old-fashioned way of running gandlf via `gandlf_configGenerator`.
@deprecated(
    "This is a deprecated way of running GanDLF. Please, use `gandlf config-generator` cli command "
    + "instead of `gandlf_configGenerator`.\n"
    + "`gandlf_configGenerator` script would be deprecated soon."
)
def old_way():
    logger_setup()
    parser = argparse.ArgumentParser(
        prog="GANDLF_ConfigGenerator",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Generate multiple GaNDLF configurations based on a single baseline GaNDLF for experimentation.\n\n"
        + copyrightMessage,
    )

    parser.add_argument(
        "-c",
        "--config",
        metavar="",
        type=str,
        help="Path to base config.",
        required=True,
    )
    parser.add_argument(
        "-s",
        "--strategy",
        metavar="",
        type=str,
        help="Config creation strategy in a yaml format.",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="",
        type=str,
        help="Path to output directory.",
        required=True,
    )

    args = parser.parse_args()

    _generate_config(args.config, args.strategy, args.output)



from pathlib import Path

import numpy as np

from panoptica import Panoptica_Evaluator


def generate_instance_segmentation(
    prediction: np.ndarray, target: np.ndarray, panoptica_config_path: str = None
) -> dict:
    """
    Evaluate a single exam using Panoptica.

    Args:
        prediction (np.ndarray): The input prediction containing objects.
        label_path (str): The path to the reference label.
        panoptica_config_path (str): The path to the Panoptica configuration file.

    Returns:
        dict: The evaluation results.
    """

    cwd = Path(__file__).parent.absolute()
    panoptica_config_path = (
        cwd / "panoptica_config_path.yaml"
        if panoptica_config_path is None
        else panoptica_config_path
    )
    evaluator = Panoptica_Evaluator.load_from_config(panoptica_config_path)

    # call evaluate
    group2result = evaluator.evaluate(prediction_arr=prediction, reference_arr=target)

    results = {k: r.to_dict() for k, r in group2result.items()}
    return results



from pydantic import BaseModel, Field, model_validator
from typing_extensions import Self, Optional


class NestedTraining(BaseModel):
    stratified: bool = Field(
        default=False,
        description="this will perform stratified k-fold cross-validation but only with offline data splitting",
    )
    testing: int = Field(
        default=-5,
        description="this controls the number of testing data folds for final model evaluation; [NOT recommended] to disable this, use '1'",
        le=10,
    )
    validation: int = Field(
        default=-5,
        description="this controls the number of validation data folds to be used for model *selection* during training (not used for back-propagation)",
    )
    proportional: Optional[bool] = Field(default=False)

    @model_validator(mode="after")
    def validate_nested_training(self) -> Self:
        if self.proportional is not None:
            self.stratified = self.proportional
        return self



import traceback
from typing import Union
import yaml
from pydantic import ValidationError

from GANDLF.configuration.parameters_config import Parameters
from GANDLF.configuration.exclude_parameters import exclude_parameters
from GANDLF.configuration.utils import handle_configuration_errors


def _parseConfig(
    config_file_path: Union[str, dict], version_check_flag: bool = True
) -> None:
    """
    This function parses the configuration file and returns a dictionary of parameters.

    Args:
        config_file_path (Union[str, dict]): The filename of the configuration file.
        version_check_flag (bool, optional): Whether to check the version in configuration file. Defaults to True.

    Returns:
        dict: The parameter dictionary.
    """
    params = config_file_path
    if not isinstance(config_file_path, dict):
        params = yaml.safe_load(open(config_file_path, "r"))

    return params


def ConfigManager(
    config_file_path: Union[str, dict], version_check_flag: bool = True
) -> dict:
    """
    This function parses the configuration file and returns a dictionary of parameters.

    Args:
        config_file_path (Union[str, dict]): The filename of the configuration file.
        version_check_flag (bool, optional): Whether to check the version in configuration file. Defaults to True.

    Returns:
        dict: The parameter dictionary.
    """
    try:
        parameters_config = Parameters(
            **_parseConfig(config_file_path, version_check_flag)
        )
        parameters = parameters_config.model_dump(
            exclude={
                field
                for field in exclude_parameters
                if getattr(parameters_config, field) is None
            }
        )
        return parameters

    except Exception as e:
        if isinstance(e, ValidationError):
            handle_configuration_errors(e)
            raise
        ## todo: ensure logging captures assertion errors
        else:
            assert (
                False
            ), f"Config parsing failed: {config_file_path=}, {version_check_flag=}, Exception: {str(e)}, {traceback.format_exc()}"
        # logging.error(
        #     f"gandlf config parsing failed: {config_file_path=}, {version_check_flag=}, Exception: {str(e)}, {traceback.format_exc()}"
        # )
        # raise



"""
All the losses are to be called from here
"""
from .segmentation import (
    MCD_loss,
    MCD_log_loss,
    MCT_loss,
    KullbackLeiblerDivergence,
    FocalLoss,
    MCC_loss,
    MCC_log_loss,
)
from .regression import CE, CEL, MSE_loss, L1_loss
from .hybrid import DCCE, DCCE_Logits, DC_Focal

# global defines for the losses
global_losses_dict = {
    "dc": MCD_loss,
    "dice": MCD_loss,
    "dc_log": MCD_log_loss,
    "dclog": MCD_log_loss,
    "dice_log": MCD_log_loss,
    "dicelog": MCD_log_loss,
    "mcc": MCC_loss,
    "mcc_log": MCC_log_loss,
    "mcclog": MCC_log_loss,
    "mathews": MCC_loss,
    "mathews_log": MCC_log_loss,
    "dcce": DCCE,
    "dcce_logits": DCCE_Logits,
    "ce": CE,
    "mse": MSE_loss,
    "cel": CEL,
    "tversky": MCT_loss,
    "kld": KullbackLeiblerDivergence,
    "l1": L1_loss,
    "focal": FocalLoss,
    "dc_focal": DC_Focal,
}


def get_loss(params: dict) -> object:
    """
    Function to get the loss definition.

    Args:
        params (dict): The parameters' dictionary.

    Returns:
        loss (object): The loss definition.
    """
    # TODO This check looks like legacy code, should we have it?

    if isinstance(params["loss_function"], dict):
        chosen_loss = list(params["loss_function"].keys())[0].lower()
    else:
        chosen_loss = params["loss_function"].lower()
    assert (
        chosen_loss in global_losses_dict
    ), f"Could not find the requested loss function '{params['loss_function']}'"

    return global_losses_dict[chosen_loss]



import torch.nn as nn
import torch.nn.functional as F


class GlobalAveragePooling2D(nn.Module):
    """
    Global average pooling operation for 2D tensors.
    """

    def __init__(self):
        super(GlobalAveragePooling2D, self).__init__()

    def forward(self, x):
        """
        Forward pass for global average pooling of a 2D tensor.

        Args:
            x (torch.Tensor):
                Input tensor with shape (batch_size, channels, height, width).

        Returns:
            torch.Tensor: Resulting tensor with shape (batch_size, channels).
        """
        assert (
            len(x.size()) == 4
        ), f"Expected 4D input tensor, but got {len(x.size())}D tensor."

        B, C, W, H = x.size()

        # This is a temporary fix to make sure the size is an integer, not a tensor
        if isinstance(B, int):
            return F.avg_pool2d(x, (W, H)).view(B, C)
        else:
            return F.avg_pool2d(x, (W.item(), H.item())).view(B.item(), C.item())


class GlobalAveragePooling3D(nn.Module):
    """
    Global average pooling operation for 3D tensors.
    """

    def __init__(self):
        super(GlobalAveragePooling3D, self).__init__()

    def forward(self, x):
        """
        Forward pass for global average pooling of a 3D tensor.

        Args:
            x (torch.Tensor): Input tensor with shape (batch_size, channels, depth, height, width).

        Returns:
            torch.Tensor: Resulting tensor with shape (batch_size, channels).
        """
        assert (
            len(x.size()) == 5
        ), f"Expected 5D input tensor, but got {len(x.size())}D tensor."
        B, C, W, H, D = x.size()

        # This is a temporary fix to make sure the size is an integer, not a tensor
        if isinstance(B, int):
            return F.avg_pool3d(x, (W, H, D)).view(B, C)
        else:
            return F.avg_pool3d(x, (W.item(), H.item(), D.item())).view(
                B.item(), C.item()
            )



from typing_extensions import Literal

from pydantic import BaseModel, Field, ConfigDict

ACCOUNTANT_OPTIONS = Literal["rdp", "gdp", "prv"]


class DifferentialPrivacyConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    noise_multiplier: float = Field(default=10.0)
    max_grad_norm: float = Field(default=1.0)
    accountant: ACCOUNTANT_OPTIONS = Field(default="rdp")
    secure_mode: bool = Field(default=False)
    allow_opacus_model_fix: bool = Field(default=True)
    delta: float = Field(default=1e-5)
    physical_batch_size: int = Field(validate_default=True)
